/*

    mii-tool: monitor and control the MII for a network interface

    Usage:

	mii-tool [-VvRrw] [-A media,... | -F media] [interface ...]

    This program is based on Donald Becker's "mii-diag" program, which
    is more capable and verbose than this tool, but also somewhat
    harder to use.

    Copyright (C) 2000 David A. Hinds -- dhinds@pcmcia.sourceforge.org

    mii-diag is written/copyright 1997-2000 by Donald Becker
        <becker@scyld.com>

    This program is free software; you can redistribute it
    and/or modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation.

    Donald Becker may be reached as becker@scyld.com, or C/O
    Scyld Computing Corporation, 410 Severn Av., Suite 210,
    Annapolis, MD 21403

    References
	http://www.scyld.com/diag/mii-status.html
	http://www.scyld.com/expert/NWay.html
	http://www.national.com/pf/DP/DP83840.html
*/

static char version[] =
"mii-tool.c 1.9 2000/04/28 00:56:08 (David Hinds)\n";

#include <sys/types.h>

#include <inttypes.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <time.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#ifndef __GLIBC__
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#endif

#include "mii.h"
#include "phyid.h"

#define MAX_ETH		8		/* Maximum # of interfaces */

typedef struct phy_info_s {
    uint32_t	phy_oui	 :PHY_VENDOR_SHIFT,	/* Vendor ID (OUI) */
		phy_model:PHY_MODEL_SHIFT;	/* Model ID/Revision */
    uint32_t	phy_bmap;			/* MDI register usage bitmap */
    const char *phy_name;			/* Model name */
} phy_info_t;

/*
 * MDI Register usage for known 8255x-compatible PHYs:
 *
 * Generic MDI registers (0-15):
 *
 * Register			       1 1 1 1 1 1
 *		   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5	 (15) Bitmap ( 0)
 * -----------------------------------------------------------------------
 * Broadcom 5221   * * * * * * * * * * * * * * *	      0x7FFF
 * Intel 82553	   * * * * * * *			      0x007F
 * Intel 82555	   * * * * * * *			      0x007F
 * National 83840  * * * * * * *			      0x007F
 * National 83843  * * * * * * * *			      0x00FF
 * National 83846  * * * * * * * *			      0x00FF
 *
 *
 * Vendor-unique MDI registers (16-31):
 *
 * Register	   1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3
 *		   6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1	 (31) Bitmap (16)
 * -----------------------------------------------------------------------
 * Broadcom 5221     * * * * * * * * * * * * * * *	      0xFFFE
 * Intel 82553	   *	   *				      0x0011
 * Intel 82555	   * * * * * * * * * * * *		      0x0FFF
 * National 83840      * * * * * * * *	 * *		      0x1CFB
 * National 83843  * * * * * * * * * *			      0x03FF
 * National 83846  *	   * * *     * * *		      0x07E1
 */

#define	PHY(oui, model, bmap, name)	\
  { PHY_OUI_##oui, PHY_MODEL_##oui##_##model, bmap, (name)}

/*
 * Known PHYs: these MUST be sorted on vendor ID and then by model ID!
 */
static phy_info_t phy_info[] = {
  PHY(AMD,	 79C901A_10BT, 0xFFFFFFFF, "AMD 79C901A 10-Base-T"),
  PHY(AMD,	 79C901A_HPNA, 0xFFFFFFFF, "AMD 79C901A Home PNA"),
  PHY(ICS,	 1889,	       0xFFFFFFFF, "ICS 1889"),
  PHY(ICS,	 1890,	       0xFFFFFFFF, "ICS 1890"),
  PHY(ICS,	 1892,	       0xFFFFFFFF, "ICS 1892"),
  PHY(ADHOC,	 AC101LF,      0xFFFFFFFF, "Altimata AC101LF"),
  PHY(ADHOC,	 AH101LF,      0xFFFFFFFF, "AdHoc AH101LF"),
  PHY(BROADCOM,	 5221,	       0xFFFE7FFF, "Broadcom BCM5221"),
  PHY(ENABLE,	 EL40331,      0xFFFFFFFF, "Enable EL40-331"),
  PHY(QUALITY,	 6612,	       0xFFFFFFFF, "Quality QS6612"),
  PHY(DAVICOM,	 DM9101,       0xFFFFFFFF, "Davicom DM9101"),
  PHY(SMSC,	 83C180,       0xFFFFFFFF, "SMSC 83C180"),
  PHY(INTEL,	 82555,	       0x0FFF007F, "Intel 82555"),
  PHY(INTEL,	 82562,	       0xFFFFFFFF, "Intel 82562ET/EM"),
  PHY(INTEL,	 82553,	       0x0011007F, "Intel 82553"),
  PHY(TDK,	 78Q2120,      0xFFFFFFFF, "TDK 78Q2120"),
  PHY(NS,	 83840,	       0x1CFB007F, "National DP83840"),
  PHY(NS,	 83843,	       0x03FF00FF, "National DP83843"),
  PHY(NS,	 83846,	       0x07E100FF, "National DP83846"),
  PHY(L1,	 LXT97x,       0xFFFFFFFF, "Level One LXT970/971")
};

#define	NMII	(sizeof (phy_info) / sizeof (phy_info[0]))

/*--------------------------------------------------------------------*/

struct option longopts[] = {
 /* { name  has_arg  *flag  val } */
    {"advertise",	1, 0, 'A'},	/* Change capabilities advertised. */
    {"force",		1, 0, 'F'},	/* Change capabilities advertised. */
    {"phy",		1, 0, 'p'},	/* Set PHY (MII address) to report. */
    {"log",		0, 0, 'l'},	/* Set PHY (MII address) to report. */
    {"restart",		0, 0, 'r'},	/* Restart link negotiation */
    {"reset",		0, 0, 'R'},	/* Reset the transceiver. */
    {"verbose", 	0, 0, 'v'},	/* Report each action taken.  */
    {"version", 	0, 0, 'V'},	/* Emit version information.  */
    {"watch", 		0, 0, 'w'},	/* Constantly monitor the port.  */
    {"help", 		0, 0, '?'},	/* Give help */
    { 0, 0, 0, 0 }
};

static unsigned int
    verbose = 0,
    opt_version = 0,
    opt_restart = 0,
    opt_reset = 0,
    opt_log = 0,
    opt_watch = 0;
static int nway_advertise = 0;
static int fixed_speed = 0;
static int override_phy = -1;

static int skfd = -1;		/* AF_INET socket for ioctl() calls. */
static struct ifreq ifr;

/*--------------------------------------------------------------------*/

static int mdio_read(int skfd, int location)
{
    struct mii_data *mii = (struct mii_data *)&ifr.ifr_data;
    mii->reg_num = location;
    if (ioctl(skfd, SIOCGMIIREG, &ifr) < 0) {
	fprintf(stderr, "SIOCGMIIREG on %s failed: %s\n", ifr.ifr_name,
		strerror(errno));
	return -1;
    }
    return mii->val_out;
}

static void mdio_write(int skfd, int location, int value)
{
    struct mii_data *mii = (struct mii_data *)&ifr.ifr_data;
    mii->reg_num = location;
    mii->val_in = value;
    if (ioctl(skfd, SIOCSMIIREG, &ifr) < 0) {
	fprintf(stderr, "SIOCSMIIREG on %s failed: %s\n", ifr.ifr_name,
		strerror(errno));
    }
}

/*--------------------------------------------------------------------*/

const struct {
    char	*name;
    u_short	value;
} media[] = {
    /* The order through 100baseT4 matches bits in the BMSR */
    { "10baseT-HD",	MII_AN_10BASET_HD },
    { "10baseT-FD",	MII_AN_10BASET_FD },
    { "100baseTx-HD",	MII_AN_100BASETX_HD },
    { "100baseTx-FD",	MII_AN_100BASETX_FD },
    { "100baseT4",	MII_AN_100BASET4 },
    { "100baseTx",	MII_AN_100BASETX_FD | MII_AN_100BASETX_HD },
    { "10baseT",	MII_AN_10BASET_FD | MII_AN_10BASET_HD },
};
#define NMEDIA (sizeof(media)/sizeof(media[0]))
	
/* Parse an argument list of media types */
static int parse_media(char *arg)
{
    int mask, i;
    char *s;
    mask = strtoul(arg, &s, 16);
    if ((*arg != '\0') && (*s == '\0')) {
	if ((mask & MII_AN_ABILITY_MASK) &&
	    !(mask & ~MII_AN_ABILITY_MASK))
	    return mask;
	goto failed;
    } else {
	mask = 0;
	s = strtok(arg, ", ");
	do {
	    for (i = 0; i < NMEDIA; i++)
		if (strcasecmp(media[i].name, s) == 0) break;
	    if (i == NMEDIA) goto failed;
	    mask |= media[i].value;
	} while ((s = strtok(NULL, ", ")) != NULL);
    }
    return mask;
failed:
    fprintf(stderr, "Invalid media specification '%s'.\n", arg);
    return -1;
}

/*--------------------------------------------------------------------*/

static char *media_list(int mask, int best)
{
    static char buf[100];
    int i;
    *buf = '\0';
    mask >>= 5;
    for (i = 4; i >= 0; i--) {
	if (mask & (1<<i)) {
	    strcat(buf, " ");
	    strcat(buf, media[i].name);
	    if (best) break;
	}
    }
    if (mask & (1<<5))
	strcat(buf, " flow-control");
    return buf;
}

static phy_info_t *
get_phy_info(uint32_t id)
{
	int min = 0, max = NMII - 1;
	uint32_t oui = PHY_VENDOR_ID(id);
	uint16_t model = PHY_MODEL(id) & 0xFFF0;
	phy_info_t *pip;

	while (1) {
		int i = (min + max) >> 1;
		int32_t order;

		pip = &phy_info[i];

		/* Attempt to match vendor */

		order = pip->phy_oui - oui;

		/* Vendor matched, search on model */

		if (order == 0) order = pip->phy_model - model;

		if (order < 0) {
			min = i + 1;

			if (min > max) return (NULL);

			continue;
		}

		if (order > 0) {
			max = i - 1;

			if (min > max) return (NULL);

			continue;
		}

		return (&phy_info[i]);
	}
}

int show_basic_mii(int sock, int phy_id)
{
    char *t, buf[100];
    uint32_t id;
    phy_info_t *pip;
    int i, mii_val[MII_MAX];
    int bmcr, bmsr, advert, lkpar;

    /* Some bits in the BMSR are latched, but we can't rely on being
       the only reader, so only the current values are meaningful */
    mdio_read(sock, MII_BMSR);
    for (i = 0; i < ((verbose > 1) ? MII_MAX : 8); i++)
	mii_val[i] = mdio_read(sock, i);

    if (mii_val[MII_BMCR] == 0xffff) {
	fprintf(stderr, "  No MII transceiver present!.\n");
	return -1;
    }

    /* Descriptive rename. */
    bmcr = mii_val[MII_BMCR]; bmsr = mii_val[MII_BMSR];
    advert = mii_val[MII_ANAR]; lkpar = mii_val[MII_ANLPAR];

    sprintf(buf, "%s: ", ifr.ifr_name);
    if (bmcr & MII_BMCR_AN_ENA) {
	if (bmsr & MII_BMSR_AN_COMPLETE) {
	    if (advert & lkpar) {
		strcat(buf, (lkpar & MII_AN_ACK) ?
		       "negotiated" : "no autonegotiation,");
		strcat(buf, media_list(advert & lkpar, 1));
		strcat(buf, ", ");
	    } else {
		strcat(buf, "autonegotiation failed, ");
	    }
	} else if (bmcr & MII_BMCR_RESTART) {
	    strcat(buf, "autonegotiation restarted, ");
	}
    } else {
	sprintf(buf+strlen(buf), "%s Mbit, %s duplex, ",
	       (bmcr & MII_BMCR_100MBIT) ? "100" : "10",
	       (bmcr & MII_BMCR_DUPLEX) ? "full" : "half");
    }
    strcat(buf, (bmsr & MII_BMSR_LINK_VALID) ? "link ok" : "no link");

    if (opt_watch) {
	if (opt_log) {
	    syslog(LOG_INFO, buf);
	} else {
	    char s[20];
	    time_t t = time(NULL);
	    strftime(s, sizeof(s), "%T", localtime(&t));
	    printf("%s %s\n", s, buf);
	}
    } else {
	printf("%s\n", buf);
    }

    /* Construct the 32-bit PHY ID and try to get known info about it */

    id = (mii_val[MII_PHY_ID1] << 16) | mii_val[MII_PHY_ID2];
    pip = get_phy_info(id);

    /*
     * Display MII registers, printing "----" for unsupported registers
     * if information about register support is available and verbose is
     * not greater than level two.
     */

    if (verbose > 1) {
	printf("  registers for MII PHY %d: ", phy_id);
	for (i = 0; i < MII_MAX; i++) {
		t = (i % 8) ? "" : "\n   ";
		if ((verbose > 2) || (pip != NULL && (1 << i) & pip->phy_bmap))
			printf("%s %4.4x", t, mii_val[i]);
		else
			printf("%s ----", t);
	}
	printf("\n");
    }

    if (verbose) {
	printf("  product info: ");
	if (pip != NULL)
	    printf("%s rev %d\n", pip->phy_name, PHY_MODEL_REV(id));
	else
	    printf("vendor %02x:%02x:%02x, model %d rev %d\n",
		   (PHY_VENDOR_ID(id) & 0xFF0000) >> 16,
		   (PHY_VENDOR_ID(id) & 0xFF00) >> 8,
		   PHY_VENDOR_ID(id) & 0xFF,
		   PHY_MODEL_ID(id), PHY_MODEL_REV(id));
	printf("  basic mode:   ");
	if (bmcr & MII_BMCR_RESET)
	    printf("software reset, ");
	if (bmcr & MII_BMCR_LOOPBACK)
	    printf("loopback, ");
	if (bmcr & MII_BMCR_ISOLATE)
	    printf("isolate, ");
	if (bmcr & MII_BMCR_COLTEST)
	    printf("collision test, ");
	if (bmcr & MII_BMCR_AN_ENA) {
	    printf("autonegotiation enabled\n");
	} else {
	    printf("%s Mbit, %s duplex\n",
		   (bmcr & MII_BMCR_100MBIT) ? "100" : "10",
		   (bmcr & MII_BMCR_DUPLEX) ? "full" : "half");
	}
	printf("  basic status: ");
	if (bmsr & MII_BMSR_AN_COMPLETE)
	    printf("autonegotiation complete, ");
	else if (bmcr & MII_BMCR_RESTART)
	    printf("autonegotiation restarted, ");
	if (bmsr & MII_BMSR_REMOTE_FAULT)
	    printf("remote fault, ");
	printf((bmsr & MII_BMSR_LINK_VALID) ? "link ok" : "no link");
	printf("\n  capabilities:%s", media_list(bmsr >> 6, 0));
	printf("\n  advertising: %s", media_list(advert, 0));
	if (lkpar & MII_AN_ABILITY_MASK)
	    printf("\n  link partner:%s", media_list(lkpar, 0));
	printf("\n");
    }
    return 0;
}

/*--------------------------------------------------------------------*/

static int do_one_xcvr(int skfd, char *ifname, int maybe)
{
    struct mii_data *mii = (struct mii_data *)&ifr.ifr_data;

    /* Get the vitals from the interface. */
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    if (ioctl(skfd, SIOCGMIIPHY, &ifr) < 0) {
	if (!maybe || (errno != ENODEV))
	    fprintf(stderr, "SIOCGMIIPHY on '%s' failed: %s\n",
		    ifname, strerror(errno));
	return 1;
    }

    if (override_phy >= 0) {
	printf("using the specified MII index %d.\n", override_phy);
	mii->phy_id = override_phy;
    }

    if (opt_reset) {
	printf("resetting the transceiver...\n");
	mdio_write(skfd, MII_BMCR, MII_BMCR_RESET);
    }
    if (nway_advertise) {
	mdio_write(skfd, MII_ANAR, nway_advertise | 1);
	opt_restart = 1;
    }
    if (opt_restart) {
	printf("restarting autonegotiation...\n");
	mdio_write(skfd, MII_BMCR, 0x0000);
	mdio_write(skfd, MII_BMCR, MII_BMCR_AN_ENA|MII_BMCR_RESTART);
    }
    if (fixed_speed) {
	int bmcr = 0;
	if (fixed_speed & (MII_AN_100BASETX_FD|MII_AN_100BASETX_HD))
	    bmcr |= MII_BMCR_100MBIT;
	if (fixed_speed & (MII_AN_100BASETX_FD|MII_AN_10BASET_FD))
	    bmcr |= MII_BMCR_DUPLEX;
	mdio_write(skfd, MII_BMCR, bmcr);
    }

    if (!opt_restart && !opt_reset && !fixed_speed && !nway_advertise)
	show_basic_mii(skfd, mii->phy_id);

    return 0;
}

/*--------------------------------------------------------------------*/

static void watch_one_xcvr(int skfd, char *ifname, int index)
{
    struct mii_data *mii = (struct mii_data *)&ifr.ifr_data;
    static int status[MAX_ETH] = { 0, /* ... */ };
    int now;

    /* Get the vitals from the interface. */
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    if (ioctl(skfd, SIOCGMIIPHY, &ifr) < 0) {
	if (errno != ENODEV)
	    fprintf(stderr, "SIOCGMIIPHY on '%s' failed: %s\n",
		    ifname, strerror(errno));
	return;
    }
    now = (mdio_read(skfd, MII_BMCR) |
	   (mdio_read(skfd, MII_BMSR) << 16));
    if (status[index] && (status[index] != now))
	show_basic_mii(skfd, mii->phy_id);
    status[index] = now;
}

/*--------------------------------------------------------------------*/

const char *usage =
"usage: %s [-VvRrwl] [-A media,... | -F media] [interface ...]\n\
       -V, --version               display version information\n\
       -v, --verbose               more verbose output\n\
       -R, --reset                 reset MII to poweron state\n\
       -r, --restart               restart autonegotiation\n\
       -w, --watch                 monitor for link status changes\n\
       -l, --log                   with -w, write events to syslog\n\
       -A, --advertise=media,...   advertise only specified media\n\
       -F, --force=media           force specified media technology\n\
media: 100baseT4, 100baseTx-FD, 100baseTx-HD, 10baseT-FD, 10baseT-HD,\n\
       (to advertise both HD and FD) 100baseTx, 10baseT\n";

int main(int argc, char **argv)
{
    int i, c, ret, errflag = 0;
    char s[6];
    
    while ((c = getopt_long(argc, argv, "A:F:p:lrRvVw?", longopts, 0)) != EOF)
	switch (c) {
	case 'A': nway_advertise = parse_media(optarg); break;
	case 'F': fixed_speed = parse_media(optarg); break;
	case 'p': override_phy = atoi(optarg); break;
	case 'r': opt_restart++;	break;
	case 'R': opt_reset++;		break;
	case 'v': verbose++;		break;
	case 'V': opt_version++;	break;
	case 'w': opt_watch++;		break;
	case 'l': opt_log++;		break;
	case '?': errflag++;
	}
    /* Check for a few inappropriate option combinations */
    if (opt_watch) verbose = 0;
    if (errflag || (fixed_speed & (fixed_speed-1)) ||
	(fixed_speed && (opt_restart || nway_advertise))) {
	fprintf(stderr, usage, argv[0]);
	return 2;
    }

    if (opt_version)
	printf(version);

    /* Open a basic socket. */
    if ((skfd = socket(AF_INET, SOCK_DGRAM,0)) < 0) {
	perror("socket");
	exit(-1);
    }

    /* No remaining args means show all interfaces. */
    if (optind == argc) {
	ret = 1;
	for (i = 0; i < MAX_ETH; i++) {
	    sprintf(s, "eth%d", i);
	    ret &= do_one_xcvr(skfd, s, 1);
	}
	if (ret)
	    fprintf(stderr, "no MII interfaces found\n");
    } else {
	ret = 0;
	for (i = optind; i < argc; i++) {
	    ret |= do_one_xcvr(skfd, argv[i], 0);
	}
    }

    if (opt_watch && (ret == 0)) {
	while (1) {
	    sleep(1);
	    if (optind == argc) {
		for (i = 0; i < MAX_ETH; i++) {
		    sprintf(s, "eth%d", i);
		    watch_one_xcvr(skfd, s, i);
		}
	    } else {
		for (i = optind; i < argc; i++)
		    watch_one_xcvr(skfd, argv[i], i-optind);
	    }
	}
    }

    close(skfd);
    return ret;
}
